import os
from datetime import datetime
from watchdog.events import FileSystemEventHandler


class Handler(FileSystemEventHandler):

    def __init__(self, queue):
        self.queue = queue

    def dispatch(self, event):
        if event.is_directory:
            return
        return super().dispatch(event)

    def on_created(self, event):
        self.queue.put(("created", event.src_path))

    def on_deleted(self, event):
        self.queue.put(("deleted", event.src_path))

    def on_modified(self, event):
        self.queue.put(("modified", event.src_path))

    def on_moved(self, event):
        self.queue.put(("moved", event.src_path, event.dest_path))


class FileSystemSource:

    def __init__(self, config):
        self.basedir = os.path.abspath(config.get('basedir', '.'))

    def walk(self, base=''):
        base = os.path.join(self.basedir, base)
        for root, dirs, files in os.walk(base):
            for name in files:
                yield os.path.relpath(os.path.join(root, name), self.basedir)

    def _open(self, filename, mode='rb'):
        f = self.open(filename, mode)
        stat = os.stat(f.fileno())
        return (f, datetime.utcfromtimestamp(stat.st_mtime), stat.st_size)

    def open(self, filename, mode='r'):
        return open(os.path.join(self.basedir, filename), mode)

    def watch(self):
        from queue import Queue
        from watchdog.observers import Observer

        queue = Queue()
        observer = Observer()
        observer.schedule(Handler(queue), self.basedir, recursive=True)
        observer.start()

        try:
            for filename in self.walk():
                yield ("created", filename)

            while True:
                event = queue.get()
                yield (event[0],) + tuple(
                    os.path.relpath(p, self.basedir) for p in event[1:])
        finally:
            observer.stop()
            observer.join()



class DryRunWriter:

    def __init__(self, config):
        pass

    def write_file(self, url, content):
        print("WRITE", url)

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass



import os
from fnmatch import fnmatch
from pathlib import PurePath
from werkzeug.wsgi import SharedDataMiddleware


class Static:

    def __init__(self, config, source):
        self.dirs = config.STATICFILES_DIRS
        self.exclude = config.STATICFILES_EXCLUDE
        self.source = source

    def is_allowed(self, path):
        return not any(
            fnmatch(s, p)
            for s in PurePath(path).parts
            for p in self.exclude)

    def __iter__(self):
        for prefix, base in self.dirs.items():
            for filename in self.source.walk(base):
                if self.is_allowed(filename):
                    name = os.path.relpath(filename, base)
                    yield f'{prefix}/{name}'

    def __call__(self, application):
        application = SharedDataMiddleware(application, self.dirs, cache=False)
        application.is_allowed = self.is_allowed
        application._opener = lambda filename: lambda: self.source._open(filename)
        return application



import json
import sqlite3
from . import Entry

def convert_json(s):
    return json.loads(s)

sqlite3.register_converter("JSON", convert_json)


class JsonGroupArray:
    def __init__(self):
        self.data = []

    def step(self, value):
        if value is None:
            return
        self.data.append(value)

    def finalize(self):
        return "[" + ",".join(self.data) + "]"


class SqliteRegistry:

    def __init__(self, config):
        conn = sqlite3.connect(
            ':memory:',
            check_same_thread=False,
            detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES)

        conn.create_aggregate("json_group_array", 1, JsonGroupArray)

        with conn:
            conn.execute(
                '''
                CREATE TABLE IF NOT EXISTS source(
                filename TEXT UNIQUE,
                reader TEXT,
                metadata JSON
                )''')

        self.conn = conn

    def __enter__(self):
        self.conn.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.conn.__exit__(exc_type, exc_value, traceback)

    def _find(self, values, args):
        args = [arg for arg in args if arg not in values]
        condition = ' AND '.join(
            ['json_extract(metadata, ?) = ?' for _ in values] +
            ['json_extract(metadata, ?) IS NOT NULL' for _ in args])

        return self.conn.execute(
            f'''SELECT oid, filename, reader, metadata FROM source WHERE {condition}''',
            tuple(p
                  for k, v in values.items()
                  for p in [f'$.{k}', v])
            + tuple(f'$.{k}' for k in args))

    def find(self, values):
        result = self._find(values, ()).fetchone()
        if result is not None:
            oid, filename, reader, metadata = result
            return Entry(metadata, oid=oid, filename=filename, reader=reader)

    def find_all(self, values, args=()):
        return [
            Entry(metadata, oid=oid, filename=filename, reader=reader)
            for oid, filename, reader, metadata in self._find(values, args).fetchall()]

    def add(self, filename, reader, metadata):
        self.conn.execute(
            '''INSERT OR REPLACE INTO source VALUES (?,?,json(?))''',
            (filename, reader, json.dumps(metadata)))

    def remove(self, filename):
        return self.conn.execute(
            '''DELETE FROM source WHERE filename = ?''',
            (filename,)).rowcount > 0



import os

class FileSystemWriter:

    def __init__(self, config):
        self.basedir = os.path.abspath(config.get('path', 'output'))

    def write_file(self, url, content):
        if url.endswith('/'):
            url += 'index.html'

        filename = os.path.join(self.basedir, url[1:])
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'wb') as f:
            f.write(content)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass



from markdown import Markdown
from markdown.preprocessors import NormalizeWhitespace
from markdown.extensions.meta import MetaPreprocessor, END_RE
from itertools import islice


def metalines(f):
    i = iter(f)
    for line in islice(i, 1):
        yield line.rstrip()

    for line in i:
        yield line.rstrip()
        if line.strip() == '' or END_RE.match(line):
            break


class MarkdownReader:

    def __init__(self, config):
        self.md = Markdown(
            extensions=config.get('extensions', ['meta']),
            extension_configs=config.get('extension_configs', {}),
            output_format=config.get("output_format", "html5"),
            tab_length=config.get("tab_length", 4))

    def metadata(self, f):
        md = self.md
        md.reset()
        lines = list(metalines(f))
        lines = NormalizeWhitespace(md).run(lines)
        lines = MetaPreprocessor(md).run(lines)
        return md.Meta

    def html(self, f):
        md = self.md
        md.reset()
        return md.convert(f.read())



from threading import Lock
from . import Entry

class DummyRegistry:

    def __init__(self, config):
        self._entries = {}
        self._lock = Lock()

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._lock.release()

    def iter(self, values, args=()):
        for entry in self._entries.values():
            if not all(
                    (k in entry) and (entry[k] == v)
                    for k, v in values.items()):
                continue
            if not all((k in entry) for k in args):
                continue
            yield entry

    def find(self, values):
        try:
            return next(self.iter(values))
        except StopIteration:
            pass

    def find_all(self, values, args=()):
        return list(self.iter(values, args))

    def add(self, filename, reader, metadata):
        self._entries[filename] = Entry(metadata, filename=filename, reader=reader)

    def remove(self, filename):
        try:
            del self._entries[filename]
            return True
        except KeyError:
            pass



import os
from types import ModuleType
from warnings import warn
from werkzeug.routing import Map, Rule
from mako.lookup import TemplateLookup
from pkg_resources import load_entry_point


class LazyDict:

    def __init__(self, func):
        self.data = {}
        self.func = func

    def __getitem__(self, key):
        if key not in self.data:
            self.data[key] = self.func(key)
        return self.data[key]


class Config:

    def __init__(self, filename, registry='default'):
        filename = os.path.abspath(filename)
        mod = ModuleType("__config__")
        mod.__file__ = filename
        mod.Map = Map
        mod.Rule = Rule
        mod.warn = warn

        with open(filename, 'r') as f:
            code = compile(f.read(), filename, 'exec')
        exec(code, mod.__dict__)
        config = mod.__dict__

        self._config = {
            'registries': config.get('REGISTRIES', {}),
            'sources': config.get('SOURCES', {}),
            'readers': config.get('READERS', {}),
            'writers': config.get('WRITERS', {}),
        }

        self._config['registries'].setdefault('default', {'name': 'sqlite'})
        self._config['sources'].setdefault('default', {'name': 'fs'})
        self._config['writers'].setdefault('default', {'name': 'dry-run'})

        self.SOURCES = LazyDict(lambda x: self.load_plugin('sources', x))
        self.READERS = LazyDict(lambda x: self.load_plugin('readers', x))
        self.WRITERS = LazyDict(lambda x: self.load_plugin('writers', x))

        self.SOURCE_FILES = config.get('SOURCE_FILES', [])

        self.STATICFILES_DIRS = config.get('STATICFILES_DIRS', {})
        self.STATICFILES_EXCLUDE = config.get('STATICFILES_EXCLUDE', [])

        self.TEMPLATES = TemplateLookup(
            config.get('TEMPLATE_DIRS', []),
            input_encoding='utf-8',
            format_exceptions=True,
            cache_enabled=False,
            imports=['from warnings import warn'])
        self.URLS = config.get('URLS', Map([]))

        mod.registry = self.registry = self.load_plugin('registries', registry)

    def load_plugin(self, type, name):
        config = self._config[type].get(name, {})
        klass = load_entry_point(
            __package__, f'{__package__}_{type}', config.get('name', name))
        return klass(config)



from werkzeug.wrappers import Response
from werkzeug.exceptions import HTTPException, NotFound

class Site:

    def __init__(self, config, source):
        self.readers = config.READERS
        self.templates = config.TEMPLATES
        self.urls = config.URLS
        self.registry = config.registry
        self.source = source

    def render(self, filename, context):
        return self.templates.get_template(filename).render(**context)

    def __iter__(self):
        urls = self.urls.bind("localhost", "/")
        for rule in self.urls.iter_rules():
            if rule.arguments:
                for entry in self.registry.find_all(rule.defaults or {}, rule.arguments or ()):
                    yield urls.build(
                        rule.endpoint,
                        {a: rule.defaults.get(a, entry.get(a, None))
                         for a in rule.arguments})
            else:
                yield urls.build(rule.endpoint)

    def __call__(self, environ, start_response):
        urls = self.urls.bind_to_environ(environ)
        try:
            endpoint, values = urls.match()
            with self.registry as registry:
                context = {'registry': registry, 'urls': urls}

                if values:
                    entry = self.registry.find(values)
                    if not entry:
                        raise NotFound()

                    context['filename'] = entry.filename
                    context['entry'] = entry
                    with self.source.open(entry.filename) as f:
                        context['content'] = self.readers[entry.reader].html(f)

                response = Response(self.render(endpoint, context), mimetype='text/html')
        except HTTPException as e:
            response = e
        return response(environ, start_response)



from collections.abc import Mapping

class Entry(Mapping):

    def __init__(self, metadata, **kwargs):
        self.__dict__.update(kwargs)
        self.__data = metadata

    def __getitem__(self, key):
        return self.__data[key]

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)



#!/usr/bin/env python3

from setuptools import setup

setup(
    name = 'elaphure',
    version = '0.0.4',

    url = 'https://github.com/bhuztez/elaphure',
    description = 'a static site generator',
    license = 'MIT',

    classifiers = [
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3 :: Only",
    ],

    author = 'bhuztez',
    author_email = 'bhuztez@gmail.com',

    packages = ['elaphure', 'elaphure.registries', 'elaphure.sources', 'elaphure.readers', 'elaphure.writers'],
    entry_points={
        'elaphure_registries':
        [ 'dummy = elaphure.registries.dummy:DummyRegistry',
          'sqlite = elaphure.registries.sqlite:SqliteRegistry'],
        'elaphure_sources':
        [ 'fs = elaphure.sources.fs:FileSystemSource' ],
        'elaphure_readers':
        [ 'markdown = elaphure.readers.markdown:MarkdownReader' ],
        'elaphure_writers':
        [ 'dry-run = elaphure.writers.dry_run:DryRunWriter',
          'fs = elaphure.writers.fs:FileSystemWriter',
          'gh-pages = elaphure.writers.gh_pages:GitHubPagesWriter' ],
    },
    install_requires = ['argh', 'Werkzeug', 'watchdog', 'Mako'],
    extras_require = {
        'markdown': ['Markdown'],
        'gh-pages': ['dulwich'],
    }
)



import os
import sys
import argh
from argh import arg

from .config import Config
from .static import Static
from .site import Site
from .generator import scan, watch

if __name__ == '__main__':
    configfile = '.elaphure'
else:
    configfile = sys.modules['__main__'].__file__


def build(writer='default', config=configfile, source='default', registry='default'):
    from warnings import catch_warnings
    from werkzeug.test import Client
    from werkzeug.wrappers import BaseResponse
    from contextlib import closing

    cfg = Config(config, registry)
    src = cfg.SOURCES[source]
    static = Static(cfg, src)
    site = Site(cfg, src)
    scan(cfg, src)
    client = Client(static(site), BaseResponse)

    with catch_warnings(record=True) as warnings:
        with cfg.WRITERS[writer] as w:
            for url in site:
                w.write_file(url, client.get(url).data)

            for url in static:
                w.write_file(url, client.get(url).data)

            for w in warnings:
                print("{}: {}".format(w.category.__name__, w.message))

            if warnings:
                quit(1)


def serve(address="0.0.0.0", port=8000, config=configfile, source='default', registry='default'):
    from werkzeug._reloader import run_with_reloader
    from werkzeug.serving import run_simple

    def inner():
        cfg = Config(config, registry)
        src = cfg.SOURCES[source]
        application = Static(cfg, src)(Site(cfg, src))
        watch(cfg, src)
        run_simple(address, port, application, use_debugger=True)

    run_with_reloader(inner)

parser = argh.ArghParser()
parser.add_commands([build, serve])
parser.dispatch()
quit()



from dulwich.repo import Repo
from dulwich.objects import Blob, Tree, Commit
from pathlib import PurePath

class GitHubPagesWriter:

    def __init__(self, config):
        self.repo = Repo(config.get('repo', '.'))
        self.branch = config.get('branch', 'gh-pages')
        self.tree = {}

    def write_file(self, url, content):
        segs = PurePath(url).parts[1:]
        if url.endswith("/"):
            segs += ("index.html",)

        tree = self.tree
        for s in segs[:-1]:
            subtree = tree.get(s, {})
            tree[s] = subtree
            tree = subtree

        blob = Blob.from_string(content)
        self.repo.object_store.add_object(blob)
        tree[segs[-1]] = blob.id

    def write_tree(self, tree):
        tree = Tree()
        for name, value in tree.items():
            if isinstance(value, dict):
                tree.add(name.encode('utf-8'), 0o040000, self.write_tree(value))
            else:
                tree.add(name.encode('utf-8'), 0o100644, value)
        self.repo.object_store.add_object(tree)
        return tree.id

    def commit(self):
        self.write_file("/.nojekyll", b'')
        tree = self.write_tree(self.tree)
        branch = "refs/heads/" + self.branch
        commit = self.repo.do_commit(message=b'generate GitHub Pages', tree=tree, ref=branch.encode())
        self.repo[branch.encode()] = commit

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.commit()



import os
from fnmatch import fnmatch
from glob import has_magic, _isrecursive, _ishidden
from werkzeug._internal import _log


def match(filename, pathname):
    if not has_magic(pathname):
        return filename == pathname
    dirname, basename = os.path.split(pathname)
    if _isrecursive(basename):
        if not dirname:
            return True
        while filename:
            if match(filename, dirname):
                return True
            filename = os.path.dirname(filename)
        return False

    if dirname:
        if not match(os.path.dirname(filename), dirname):
            return False
    elif os.path.dirname(filename):
        return False
    filename = os.path.basename(filename)
    if _ishidden(filename):
        return False
    return fnmatch(filename, basename)


class Generator:

    def __init__(self, config, source):
        self.files = config.SOURCE_FILES
        self.readers = config.READERS
        self.registry = config.registry
        self.source = source

    def add(self, filename):
        for pattern, reader_name, meta_func in self.files:
            if match(filename, pattern):
                reader = self.readers[reader_name]
                with self.source.open(filename) as f:
                    data = reader.metadata(f)
                self.registry.add(filename, reader_name, meta_func(filename, data))
                return True

    def remove(self, filename):
        return self.registry.remove(filename)

    def scan(self):
        with self.registry:
            for filename in self.source.walk():
                if self.add(filename):
                    _log("info", " * File found %r" % (filename,))

    def watch(self):
        for event, src_path, *dest_path in self.source.watch():
            with self.registry:
                if event == 'created':
                    if self.add(src_path):
                        _log("info", " * File found %r" % (src_path,))
                elif event == 'modified':
                    if self.add(src_path):
                        _log("info", " * File modified %r" % (src_path,))
                elif event == 'deleted':
                    if self.remove(src_path):
                        _log("info", " * File deleted %r" % (src_path,))
                elif event == 'moved':
                    remove = self.remove(src_path)
                    add = self.add(dest_path[0])
                    if remove or add:
                        _log("info", " * File moved from %r to %r" % (src_path, dest_path))

def scan(config, src):
    Generator(config, src).scan()

def watch(config, src):
    from threading import Thread
    Thread(target=Generator(config, src).watch, daemon=True).start()
